@@ -103,6 +103,7 @@ typedef enum {
 	lat_file_unlink,
 	lat_file_open,
 	lat_file_read,
+	lat_file_seek,
 	lat_file_write,
 	lat_commit,
 	lat_file_close,
@@ -122,12 +123,12 @@ struct latency_data {
 #endif
 
 struct glusterfs_fsal_module {
-	struct fsal_staticfsinfo_t fs_info;
 	struct fsal_module fsal;
+	struct fsal_obj_ops handle_ops;
 	struct glist_head  fs_obj; /* list of glusterfs_fs filesystem objects */
 	pthread_mutex_t   lock; /* lock to protect above list */
 };
-struct glusterfs_fsal_module GlusterFS;
+extern struct glusterfs_fsal_module GlusterFS;
 
 struct glusterfs_fs {
 	struct glist_head fs_obj; /* link to glusterfs_fs filesystem objects */
@@ -247,8 +248,6 @@ fsal_status_t gluster2fsal_error(const int gluster_errorcode);
 void stat2fsal_attributes(const struct stat *buffstat,
 			  struct attrlist *fsalattr);
 
-struct fsal_staticfsinfo_t *gluster_staticinfo(struct fsal_module *hdl);
-
 void construct_handle(struct glusterfs_export *glexport, const struct stat *st,
 		      struct glfs_object *glhandle, unsigned char *globjhdl,
 		      int len, struct glusterfs_handle **obj,
