@@ -18,9 +18,7 @@
 
 #include "fsal.h"
 #if 0
-#include "fsal_internal.h"
 #include "fsal_convert.h"
-#include "fsal_common.h"
 #endif
 
 #ifndef ARRAY_SIZE
@@ -491,8 +489,8 @@ do { \
 
 #define COMPOUNDV4_EXECUTE_SIMPLE(pcontext, argcompound, rescompound)   \
 	  clnt_call(pcontext->rpc_client, NFSPROC4_COMPOUND,		\
-		    (xdrproc_t)xdr_COMPOUND4args, (caddr_t)&argcompound, \
-		    (xdrproc_t)xdr_COMPOUND4res,  (caddr_t)&rescompound, \
+		    (xdrproc_t)xdr_COMPOUND4args, &argcompound,		\
+		    (xdrproc_t)xdr_COMPOUND4res,  &rescompound,		\
 		    timeout)
 
 #endif				/* _FSAL_NFSV4_MACROS_H */
