@@ -508,12 +508,14 @@ mdcache_lru_clean(mdcache_entry_t *entry)
 		/* Find the first export id. */
 		export_id = atomic_fetch_int32_t(&entry->first_export_id);
 
-		if (export_id >= 0 && op_ctx != NULL &&
-		    op_ctx->ctx_export != NULL &&
-		    op_ctx->ctx_export->export_id != export_id) {
-			/* We can't be sure the op_ctx has a valid export_id for
-			 * this entry, so we'll use the first export_id and set
-			 * up a new op_ctx.
+		/* Check if we have a valid op_ctx */
+		if (export_id >= 0 && (op_ctx == NULL ||
+		    op_ctx->ctx_export == NULL ||
+		    op_ctx->ctx_export->export_id != export_id)) {
+			/* If the entry's first_export_id is valid and does not
+			 * match the current op_ctx, set up a new context
+			 * using first_export_id to ensure the op_ctx export is
+			 * valid for the entry.
 			 *
 			 * Get a reference to the first_export_id.
 			 */
@@ -569,7 +571,7 @@ mdcache_lru_clean(mdcache_entry_t *entry)
 		}
 
 		subcall(
-			entry->sub_handle->obj_ops.release(entry->sub_handle)
+			entry->sub_handle->obj_ops->release(entry->sub_handle)
 		       );
 		entry->sub_handle = NULL;
 
@@ -585,18 +587,23 @@ mdcache_lru_clean(mdcache_entry_t *entry)
 	/* Done with the attrs */
 	fsal_release_attrs(&entry->attrs);
 
-	/* Clean our handle */
-	fsal_obj_handle_fini(&entry->obj_handle);
-
 	/* Clean out the export mapping before deconstruction */
 	mdc_clean_entry(entry);
 
+	/* Clean our handle */
+	fsal_obj_handle_fini(&entry->obj_handle);
+
 	/* Finalize last bits of the cache entry, delete the key if any and
 	 * destroy the rw locks.
 	 */
 	mdcache_key_delete(&entry->fh_hk.key);
 	PTHREAD_RWLOCK_destroy(&entry->content_lock);
 	PTHREAD_RWLOCK_destroy(&entry->attr_lock);
+
+	state_hdl_cleanup(entry->obj_handle.state_hdl);
+
+	if (entry->obj_handle.type == DIRECTORY)
+		pthread_spin_destroy(&entry->fsobj.fsdir.spin);
 }
 
 /**
@@ -643,6 +650,10 @@ lru_reap_impl(enum lru_q_id qid)
 		}
 		refcnt = atomic_inc_int32_t(&lru->refcnt);
 		entry = container_of(lru, mdcache_entry_t, lru);
+#ifdef USE_LTTNG
+	tracepoint(mdcache, mdc_lru_ref,
+		   __func__, __LINE__, entry, entry->sub_handle, refcnt);
+#endif
 		QUNLOCK(qlane);
 
 		if (unlikely(refcnt != (LRU_SENTINEL_REFCOUNT + 1))) {
@@ -735,8 +746,9 @@ lru_try_reap_entry(void)
  *
  * @note The caller @a MUST @a NOT hold the lane lock
  *
- * @param[in] qid     Queue to reap
- * @param[in] parent  The directory we desire a chunk for
+ * @param[in] qid        Queue to reap
+ * @param[in] parent     The directory we desire a chunk for
+ * @param[in] prev_chunk If non-NULL, the previous chunk in this directory
  *
  * @return Available chunk if found, NULL otherwise
  */
@@ -744,7 +756,8 @@ lru_try_reap_entry(void)
 static uint32_t chunk_reap_lane;
 
 static inline mdcache_lru_t *
-lru_reap_chunk_impl(enum lru_q_id qid, mdcache_entry_t *parent)
+lru_reap_chunk_impl(enum lru_q_id qid, mdcache_entry_t *parent,
+		    struct dir_chunk *prev_chunk)
 {
 	uint32_t lane;
 	struct lru_q_lane *qlane;
@@ -779,6 +792,12 @@ lru_reap_chunk_impl(enum lru_q_id qid, mdcache_entry_t *parent)
 		chunk = container_of(lru, struct dir_chunk, chunk_lru);
 		entry = chunk->parent;
 
+		if (chunk == prev_chunk) {
+			/* We can't reap prev_chunk. */
+			QUNLOCK(qlane);
+			continue;
+		}
+
 		/* If this chunk belongs to the parent seeking another chunk,
 		 * or if we can get the content_lock for the chunk's parent,
 		 * we can reap this chunk.
@@ -820,7 +839,6 @@ lru_reap_chunk_impl(enum lru_q_id qid, mdcache_entry_t *parent)
 
 			/* Clean out the fields not touched by the cleanup. */
 			chunk->parent = NULL;
-			chunk->prev_chunk = NULL;
 			chunk->next_ck = 0;
 			chunk->num_entries = 0;
 
@@ -850,23 +868,27 @@ lru_reap_chunk_impl(enum lru_q_id qid, mdcache_entry_t *parent)
  * @brief Re-use or allocate a chunk
  *
  * This function repurposes a resident chunk in the LRU system if the system is
- * above the high-water mark, and allocates a new one otherwise.
+ * above the high-water mark, and allocates a new one otherwise.  The resulting
+ * chunk is inserted into the chunk list.
  *
  * @note The caller must hold the content_lock of the parent for write.
  *
- * @param[in] parent  The parent directory we desire a chunk for
+ * @param[in] parent     The parent directory we desire a chunk for
+ * @param[in] prev_chunk If non-NULL, the previous chunk in this directory
  *
  * @return reused or allocated chunk
  */
-struct dir_chunk *mdcache_get_chunk(mdcache_entry_t *parent)
+struct dir_chunk *mdcache_get_chunk(mdcache_entry_t *parent,
+				    struct dir_chunk *prev_chunk)
 {
 	mdcache_lru_t *lru = NULL;
 	struct dir_chunk *chunk = NULL;
 
 	if (lru_state.chunks_used >= lru_state.chunks_hiwat) {
-		lru = lru_reap_chunk_impl(LRU_ENTRY_L2, parent);
+		lru = lru_reap_chunk_impl(LRU_ENTRY_L2, parent, prev_chunk);
 		if (!lru)
-			lru = lru_reap_chunk_impl(LRU_ENTRY_L1, parent);
+			lru = lru_reap_chunk_impl(
+					LRU_ENTRY_L1, parent, prev_chunk);
 	}
 
 	if (lru) {
@@ -885,8 +907,13 @@ struct dir_chunk *mdcache_get_chunk(mdcache_entry_t *parent)
 		(void) atomic_inc_int64_t(&lru_state.chunks_used);
 	}
 
-	/* Set the chunk's parent. */
+	/* Set the chunk's parent and insert */
 	chunk->parent = parent;
+	if (prev_chunk) {
+		glist_add(&prev_chunk->chunks, &chunk->chunks);
+	} else {
+		glist_add(&chunk->parent->fsobj.fsdir.chunks, &chunk->chunks);
+	}
 
 	/* Chunk refcnt is not used (chunks are always protected by content_lock
 	 * and outside of LRU operations are not found other than while holding
@@ -1064,9 +1091,9 @@ static inline size_t lru_run_lane(size_t lane, uint64_t *const totalclosed)
 			goto next_lane;
 
 		lru = glist_entry(qlane->iter.glist, mdcache_lru_t, q);
-		refcnt = atomic_inc_int32_t(&lru->refcnt);
 
-		/* get entry early */
+		/* get entry early.  This is safe without a ref, because we have
+		 * the QLANE lock */
 		entry = container_of(lru, mdcache_entry_t, lru);
 
 		/* Get a reference to the first export and build an op context
@@ -1079,34 +1106,34 @@ static inline size_t lru_run_lane(size_t lane, uint64_t *const totalclosed)
 		export_id = atomic_fetch_int32_t(&entry->first_export_id);
 
 		if (export_id < 0) {
-			/* This should never happen, since any entry that only
-			 * had a sentinel reference must either have a mapped
-			 * export or be in the LRU_ENTRY_CLEANUP queue and thus
-			 * not eligible for lru_run_lane.
-			 */
-			LogFatal(COMPONENT_CACHE_INODE,
-				 "No first_export for entry %p is unexpected.",
-				 entry);
+			/* This entry is part of an export that's going away.
+			 * Just skip it. */
+			continue;
 		}
 
 		export = get_gsh_export(export_id);
 
 		if (export == NULL) {
-			/* This really should not happen, if an unexport is in
-			 * progress, the export_id is now not removed until
-			 * after mdcache has detached all entries from the
-			 * export. An entry that is actually in the process of
-			 * being detached has an LRU reference which prevents it
-			 * from being processed by lru_run_lane, so there is no
-			 * path to get here without the export still being
-			 * valid.
+			/* Creating the root object of an export and inserting
+			 * the export are not atomic.  That is, we create the
+			 * root object (and it's inserted in the LRU), and then
+			 * we insert the export, to make it reachable.  This
+			 * creates a tiny window the root object is in the LRU
+			 * (and therefore visible in this function) but the
+			 * export is not yet inserted, and so the above lookup
+			 * will fail.  Skip such entries, as this is a
+			 * self-correcting situation.
 			 */
-			LogFatal(COMPONENT_CACHE_INODE,
-				 "An entry (%p) having an unmappable export_id (%"
-				 PRIi32") is unexpected",
-				 entry, export_id);
+			continue;
 		}
 
+		/* Get a ref on the entry now */
+		refcnt = atomic_inc_int32_t(&entry->lru.refcnt);
+#ifdef USE_LTTNG
+	tracepoint(mdcache, mdc_lru_ref,
+		   __func__, __LINE__, entry, entry->sub_handle, refcnt);
+#endif
+
 		init_root_op_context(&ctx, export, export->fsal_export, 0, 0,
 				     UNKNOWN_REQUEST);
 
@@ -1224,7 +1251,7 @@ lru_run(struct fridgethr_context *ctx)
 	/* Finalized */
 	uint32_t fdratepersec = 1, fds_avg, fddelta;
 	float fdnorm, fdwait_ratio, fdmulti;
-	time_t threadwait = fridgethr_getwait(ctx);
+	time_t threadwait = mdcache_param.lru_run_interval;
 	/* True if we are taking extreme measures to reclaim FDs */
 	bool extremis = false;
 	/* Total work done in all passes so far.  If this exceeds the
@@ -1240,9 +1267,7 @@ lru_run(struct fridgethr_context *ctx)
 
 	fds_avg = (lru_state.fds_hiwat - lru_state.fds_lowat) / 2;
 
-	if (mdcache_param.use_fd_cache)
-		extremis = (atomic_fetch_size_t(&open_fd_count) >
-			    lru_state.fds_hiwat);
+	extremis = atomic_fetch_size_t(&open_fd_count) > lru_state.fds_hiwat;
 
 	LogFullDebug(COMPONENT_CACHE_INODE_LRU, "LRU awakes.");
 
@@ -1250,6 +1275,10 @@ lru_run(struct fridgethr_context *ctx)
 		/* If we make it all the way through a timed sleep
 		   without being woken, we assume we aren't racing
 		   against the impossible. */
+		if (lru_state.futility >= mdcache_param.futility_count)
+			LogInfo(COMPONENT_CACHE_INODE_LRU,
+				"Leaving FD futility mode.");
+
 		lru_state.futility = 0;
 	}
 
@@ -1262,17 +1291,16 @@ lru_run(struct fridgethr_context *ctx)
 	   API, for example.) */
 
 	currentopen = atomic_fetch_size_t(&open_fd_count);
-	if ((currentopen < lru_state.fds_lowat)
-	    && mdcache_param.use_fd_cache) {
+
+	if (currentopen < lru_state.fds_lowat) {
 		LogDebug(COMPONENT_CACHE_INODE_LRU,
 			 "FD count is %zd and low water mark is %d: not reaping.",
 			 atomic_fetch_size_t(&open_fd_count),
 			 lru_state.fds_lowat);
-		if (mdcache_param.use_fd_cache
-		    && !lru_state.caching_fds) {
-			lru_state.caching_fds = true;
+		if (atomic_fetch_uint32_t(&lru_state.fd_state) > FD_LOW) {
 			LogEvent(COMPONENT_CACHE_INODE_LRU,
-				 "Re-enabling FD cache.");
+				 "Return to normal fd reaping.");
+			atomic_store_uint32_t(&lru_state.fd_state, FD_LOW);
 		}
 	} else {
 		/* The count of open file descriptors before this run
@@ -1284,7 +1312,19 @@ lru_run(struct fridgethr_context *ctx)
 		size_t workpass = 0;
 		time_t curr_time = time(NULL);
 
-		fdratepersec = (curr_time <= lru_state.prev_time)
+		if (currentopen < lru_state.fds_hiwat &&
+		    atomic_fetch_uint32_t(&lru_state.fd_state) == FD_LIMIT) {
+			LogEvent(COMPONENT_CACHE_INODE_LRU,
+				 "Count of fd is below high water mark.");
+			atomic_store_uint32_t(&lru_state.fd_state, FD_MIDDLE);
+		}
+
+		if ((curr_time >= lru_state.prev_time) &&
+		    (curr_time - lru_state.prev_time < fridgethr_getwait(ctx)))
+			threadwait = curr_time - lru_state.prev_time;
+
+		fdratepersec = ((curr_time <= lru_state.prev_time) ||
+				(formeropen < lru_state.prev_fd_count))
 			? 1 : (formeropen - lru_state.prev_fd_count) /
 					(curr_time - lru_state.prev_time);
 
@@ -1325,11 +1365,10 @@ lru_run(struct fridgethr_context *ctx)
 			       lru_state.fds_hiwat) *
 			      mdcache_param.required_progress) /
 			     100)))) {
-			if (++lru_state.futility >
+			if (++lru_state.futility ==
 			    mdcache_param.futility_count) {
-				LogCrit(COMPONENT_CACHE_INODE_LRU,
-					"Futility count exceeded.  The LRU thread is unable to make progress in reclaiming FDs.  Disabling FD cache.");
-				lru_state.caching_fds = false;
+				LogWarn(COMPONENT_CACHE_INODE_LRU,
+					"Futility count exceeded.  Client load is opening FDs faster than the LRU thread can close them.");
 			}
 		}
 	}
@@ -1365,10 +1404,10 @@ lru_run(struct fridgethr_context *ctx)
 
 	LogDebug(COMPONENT_CACHE_INODE_LRU,
 		 "After work, open_fd_count:%zd  count:%" PRIu64
-		 " fdrate:%u threadwait=%" PRIu64,
+		 " fdrate:%u new_thread_wait=%" PRIu64,
 		 atomic_fetch_size_t(&open_fd_count),
 		 lru_state.entries_used, fdratepersec,
-		 ((uint64_t) threadwait));
+		 ((uint64_t) new_thread_wait));
 	LogFullDebug(COMPONENT_CACHE_INODE_LRU,
 		     "currentopen=%zd futility=%d totalwork=%zd biggest_window=%d extremis=%d lanes=%d fds_lowat=%d ",
 		     currentopen, lru_state.futility, totalwork,
@@ -1421,7 +1460,7 @@ static inline size_t chunk_lru_run_lane(size_t lane)
 
 		/* Move lru object to MRU of L2 */
 		q = &qlane->L1;
-		LRU_DQ_SAFE(lru, q);
+		CHUNK_LRU_DQ_SAFE(lru, q);
 		lru->qid = LRU_ENTRY_L2;
 		q = &qlane->L2;
 		lru_insert(lru, q, LRU_MRU);
@@ -1616,7 +1655,7 @@ mdcache_lru_pkginit(void)
 
 	atomic_store_size_t(&open_fd_count, 0);
 	lru_state.prev_fd_count = 0;
-	lru_state.caching_fds = mdcache_param.use_fd_cache;
+	atomic_store_uint32_t(&lru_state.fd_state, FD_LOW);
 	init_fds_limit();
 
 	/* Set high and low watermark for cache entries.  XXX This seems a
@@ -1718,7 +1757,7 @@ mdcache_entry_t *alloc_cache_entry(void)
  *
  * @return a usable entry or NULL if unexport is in progress.
  */
-mdcache_entry_t *mdcache_lru_get(void)
+mdcache_entry_t *mdcache_lru_get(struct fsal_obj_handle *sub_handle)
 {
 	mdcache_lru_t *lru;
 	mdcache_entry_t *nentry = NULL;
@@ -1727,8 +1766,6 @@ mdcache_entry_t *mdcache_lru_get(void)
 	if (lru) {
 		/* we uniquely hold entry */
 		nentry = container_of(lru, mdcache_entry_t, lru);
-		LogFullDebug(COMPONENT_CACHE_INODE_LRU,
-			     "Recycling entry at %p.", nentry);
 		mdcache_lru_clean(nentry);
 		memset(&nentry->attrs, 0, sizeof(nentry->attrs));
 		init_rw_locks(nentry);
@@ -1741,26 +1778,36 @@ mdcache_entry_t *mdcache_lru_get(void)
 	nentry->lru.refcnt = 2;
 	nentry->lru.cf = 0;
 	nentry->lru.lane = lru_lane_of(nentry);
+	nentry->sub_handle = sub_handle;
 
 #ifdef USE_LTTNG
 	tracepoint(mdcache, mdc_lru_get,
-		   __func__, __LINE__, nentry, nentry->lru.refcnt);
+		  __func__, __LINE__, nentry, sub_handle, nentry->lru.refcnt);
 #endif
-
 	return nentry;
 }
 
 /**
  * @brief Insert a new entry into the LRU.
  *
- * Entry is inserted into LRU of L1 queue.
+ * Entry is inserted the LRU.  For scans, insert into the MRU of L2, to avoid
+ * having entries recycled before they're used during readdir.  For everything
+ * else, insert into LRU of L1, so that a single ref promotes to the MRU of L1.
  *
- * @param [in] ntry  Entry to insert.
+ * @param [in] entry  Entry to insert.
+ * @param [in] reason Reason we're inserting
  */
-void mdcache_lru_insert(mdcache_entry_t *entry)
+void mdcache_lru_insert(mdcache_entry_t *entry, mdc_reason_t reason)
 {
 	/* Enqueue. */
-	lru_insert_entry(entry, &LRU[entry->lru.lane].L1, LRU_LRU);
+	switch (reason) {
+	case MDC_REASON_DEFAULT:
+		lru_insert_entry(entry, &LRU[entry->lru.lane].L1, LRU_LRU);
+		break;
+	case MDC_REASON_SCAN:
+		lru_insert_entry(entry, &LRU[entry->lru.lane].L2, LRU_MRU);
+		break;
+	}
 }
 
 /**
@@ -1796,7 +1843,7 @@ _mdcache_lru_ref(mdcache_entry_t *entry, uint32_t flags, const char *func,
 
 #ifdef USE_LTTNG
 	tracepoint(mdcache, mdc_lru_ref,
-		   func, line, entry, refcnt);
+		   func, line, entry, entry->sub_handle, refcnt);
 #endif
 
 	/* adjust LRU on initial refs */
@@ -1878,7 +1925,7 @@ _mdcache_lru_unref(mdcache_entry_t *entry, uint32_t flags, const char *func,
 
 #ifdef USE_LTTNG
 	tracepoint(mdcache, mdc_lru_unref,
-		   func, line, entry, refcnt);
+		   func, line, entry, entry->sub_handle, refcnt);
 #endif
 
 	if (unlikely(refcnt == 0)) {
@@ -1915,11 +1962,11 @@ _mdcache_lru_unref(mdcache_entry_t *entry, uint32_t flags, const char *func,
 }
 
 /**
- * @brief Remove a chunk from LRU, clean it, and free it.
+ * @brief Remove a chunk from LRU, and clean it
  *
  * @param[in] chunk  The chunk to be removed from LRU
  */
-void lru_remove_chunk(struct dir_chunk *chunk)
+void lru_clean_chunk(struct dir_chunk *chunk)
 {
 	uint32_t lane = chunk->chunk_lru.lane;
 	struct lru_q_lane *qlane = &CHUNK_LRU[lane];
@@ -1943,12 +1990,37 @@ void lru_remove_chunk(struct dir_chunk *chunk)
 
 	/* Then do the actual cleaning work. */
 	mdcache_clean_dirent_chunk(chunk);
+}
+
+/**
+ * @brief Remove a chunk from LRU, clean it, and free it.
+ *
+ * @param[in] chunk  The chunk to be removed from LRU
+ */
+void lru_remove_chunk(struct dir_chunk *chunk)
+{
+	lru_clean_chunk(chunk);
 
 	/* And now we can free the chunk. */
 	LogFullDebug(COMPONENT_CACHE_INODE, "Freeing chunk %p", chunk);
 	gsh_free(chunk);
 }
 
+/**
+ * @brief Remove a chunk from LRU, and clean it
+ *
+ * @param[in] chunk  The chunk to be removed from LRU
+ */
+void lru_reuse_chunk(mdcache_entry_t *parent, struct dir_chunk *chunk)
+{
+	lru_clean_chunk(chunk);
+	chunk->parent = parent;
+	chunk->chunk_lru.refcnt = 0;
+	chunk->chunk_lru.cf = 0;
+	chunk->chunk_lru.lane = lru_lane_of(chunk);
+	lru_insert_chunk(chunk, &CHUNK_LRU[chunk->chunk_lru.lane].L2, LRU_MRU);
+}
+
 /**
  * @brief Indicate that a chunk is being used, bump it up in the LRU
  *
@@ -1957,14 +2029,15 @@ void lru_bump_chunk(struct dir_chunk *chunk)
 {
 	mdcache_lru_t *lru = &chunk->chunk_lru;
 	struct lru_q_lane *qlane = &CHUNK_LRU[lru->lane];
-	struct lru_q *q = chunk_lru_queue_of(chunk);
+	struct lru_q *q;
 
 	QLOCK(qlane);
+	q = chunk_lru_queue_of(chunk);
 
 	switch (lru->qid) {
 	case LRU_ENTRY_L1:
 		/* advance chunk to MRU (of L1) */
-		LRU_DQ_SAFE(lru, q);
+		CHUNK_LRU_DQ_SAFE(lru, q);
 		lru_insert(lru, q, LRU_MRU);
 		break;
 	case LRU_ENTRY_L2:
@@ -1983,17 +2056,40 @@ void lru_bump_chunk(struct dir_chunk *chunk)
 }
 
 /**
+ * @brief Check if FDs are available.
  *
- * @brief Wake the LRU thread to free FDs.
+ * This function checks if FDs are available to serve open
+ * requests. This function also wakes the LRU thread if the
+ * current FD count is above the high water mark.
  *
- * This function wakes the LRU reaper thread to free FDs and should be
- * called when we are over the high water mark.
+ * @return true if there are FDs available to serve open requests,
+ * false otherwise.
  */
-
-void
-lru_wake_thread(void)
+bool mdcache_lru_fds_available(void)
 {
-	fridgethr_wake(lru_fridge);
+	if (atomic_fetch_size_t(&open_fd_count) >= lru_state.fds_hard_limit) {
+		LogAtLevel(COMPONENT_CACHE_INODE_LRU,
+			   atomic_fetch_uint32_t(&lru_state.fd_state)
+								!= FD_LIMIT
+				? NIV_CRIT
+				: NIV_DEBUG,
+			   "FD Hard Limit Exceeded, waking LRU thread.");
+		atomic_store_uint32_t(&lru_state.fd_state, FD_LIMIT);
+		fridgethr_wake(lru_fridge);
+		return false;
+	}
+
+	if (atomic_fetch_size_t(&open_fd_count) >= lru_state.fds_hiwat) {
+		LogAtLevel(COMPONENT_CACHE_INODE_LRU,
+			   atomic_fetch_uint32_t(&lru_state.fd_state) == FD_LOW
+				? NIV_INFO
+				: NIV_DEBUG,
+			   "FDs above high water mark, waking LRU thread.");
+		atomic_store_uint32_t(&lru_state.fd_state, FD_HIGH);
+		fridgethr_wake(lru_fridge);
+	}
+
+	return true;
 }
 
 /** @} */
