@@ -54,6 +54,7 @@ struct nfs4_op_desc {
 		     struct nfs_resop4 *);
 
 	void (*free_res)(nfs_resop4 *);
+	uint32_t resp_size;
 	int exp_perm_flags;
 };
 
@@ -67,386 +68,462 @@ static const struct nfs4_op_desc optabv4[] = {
 		.name = "OP_ILLEGAL",
 		.funct = nfs4_op_illegal,
 		.free_res = nfs4_op_illegal_Free,
+		.resp_size = sizeof(ILLEGAL4res),
 		.exp_perm_flags = 0},
 	[1] = {
 		.name = "OP_ILLEGAL",
 		.funct = nfs4_op_illegal,
 		.free_res = nfs4_op_illegal_Free,
+		.resp_size = sizeof(ILLEGAL4res),
 		.exp_perm_flags = 0},
 	[2] = {
 		.name = "OP_ILLEGAL",
 		.funct = nfs4_op_illegal,
 		.free_res = nfs4_op_illegal_Free,
+		.resp_size = sizeof(ILLEGAL4res),
 		.exp_perm_flags = 0},
 	[NFS4_OP_ACCESS] = {
 		.name = "OP_ACCESS",
 		.funct = nfs4_op_access,
 		.free_res = nfs4_op_access_Free,
+		.resp_size = sizeof(ACCESS4res),
 		.exp_perm_flags = EXPORT_OPTION_MD_READ_ACCESS},
 	[NFS4_OP_CLOSE] = {
 		.name = "OP_CLOSE",
 		.funct = nfs4_op_close,
 		.free_res = nfs4_op_close_Free,
+		.resp_size = sizeof(CLOSE4res),
 		.exp_perm_flags = EXPORT_OPTION_MD_READ_ACCESS},
 	[NFS4_OP_COMMIT] = {
 		.name = "OP_COMMIT",
 		.funct = nfs4_op_commit,
 		.free_res = nfs4_op_commit_Free,
+		.resp_size = sizeof(COMMIT4res),
 		.exp_perm_flags = EXPORT_OPTION_MD_WRITE_ACCESS},
 	[NFS4_OP_CREATE] = {
 		.name = "OP_CREATE",
 		.funct = nfs4_op_create,
 		.free_res = nfs4_op_create_Free,
+		.resp_size = sizeof(CREATE4res),
 		.exp_perm_flags = EXPORT_OPTION_MD_WRITE_ACCESS},
 	[NFS4_OP_DELEGPURGE] = {
 		.name = "OP_DELEGPURGE",
 		.funct = nfs4_op_delegpurge,
 		.free_res = nfs4_op_delegpurge_Free,
+		.resp_size = sizeof(DELEGPURGE4res),
 		.exp_perm_flags = 0},
 	[NFS4_OP_DELEGRETURN] = {
 		.name = "OP_DELEGRETURN",
 		.funct = nfs4_op_delegreturn,
 		.free_res = nfs4_op_delegreturn_Free,
+		.resp_size = sizeof(DELEGRETURN4res),
 		.exp_perm_flags = EXPORT_OPTION_MD_READ_ACCESS},
 	[NFS4_OP_GETATTR] = {
 		.name = "OP_GETATTR",
 		.funct = nfs4_op_getattr,
 		.free_res = nfs4_op_getattr_Free,
+		.resp_size = VARIABLE_RESP_SIZE,
 		.exp_perm_flags = EXPORT_OPTION_MD_READ_ACCESS},
 	[NFS4_OP_GETFH] = {
 		.name = "OP_GETFH",
 		.funct = nfs4_op_getfh,
 		.free_res = nfs4_op_getfh_Free,
+		.resp_size = VARIABLE_RESP_SIZE,
 		.exp_perm_flags = 0},
 	[NFS4_OP_LINK] = {
 		.name = "OP_LINK",
 		.funct = nfs4_op_link,
 		.free_res = nfs4_op_link_Free,
+		.resp_size = sizeof(LINK4res),
 		.exp_perm_flags = EXPORT_OPTION_MD_WRITE_ACCESS},
 	[NFS4_OP_LOCK] = {
 		.name = "OP_LOCK",
 		.funct = nfs4_op_lock,
 		.free_res = nfs4_op_lock_Free,
+		.resp_size = VARIABLE_RESP_SIZE,
 		.exp_perm_flags = EXPORT_OPTION_MD_READ_ACCESS},
 	[NFS4_OP_LOCKT] = {
 		.name = "OP_LOCKT",
 		.funct = nfs4_op_lockt,
 		.free_res = nfs4_op_lockt_Free,
+		.resp_size = VARIABLE_RESP_SIZE,
 		.exp_perm_flags = EXPORT_OPTION_MD_READ_ACCESS},
 	[NFS4_OP_LOCKU] = {
 		.name = "OP_LOCKU",
 		.funct = nfs4_op_locku,
 		.free_res = nfs4_op_locku_Free,
+		.resp_size = sizeof(LOCKU4res),
 		.exp_perm_flags = EXPORT_OPTION_MD_READ_ACCESS},
 	[NFS4_OP_LOOKUP] = {
 		.name = "OP_LOOKUP",
 		.funct = nfs4_op_lookup,
 		.free_res = nfs4_op_lookup_Free,
+		.resp_size = sizeof(LOOKUP4res),
 		.exp_perm_flags = EXPORT_OPTION_MD_READ_ACCESS},
 	[NFS4_OP_LOOKUPP] = {
 		.name = "OP_LOOKUPP",
 		.funct = nfs4_op_lookupp,
 		.free_res = nfs4_op_lookupp_Free,
+		.resp_size = sizeof(LOOKUPP4res),
 		.exp_perm_flags = EXPORT_OPTION_MD_READ_ACCESS},
 	[NFS4_OP_NVERIFY] = {
 		.name = "OP_NVERIFY",
 		.funct = nfs4_op_nverify,
 		.free_res = nfs4_op_nverify_Free,
+		.resp_size = sizeof(NVERIFY4res),
 		.exp_perm_flags = EXPORT_OPTION_MD_READ_ACCESS},
 	[NFS4_OP_OPEN] = {
 		.name = "OP_OPEN",
 		.funct = nfs4_op_open,
 		.free_res = nfs4_op_open_Free,
+		.resp_size = sizeof(OPEN4res),
 		.exp_perm_flags = EXPORT_OPTION_MD_READ_ACCESS},
 	[NFS4_OP_OPENATTR] = {
 		.name = "OP_OPENATTR",
 		.funct = nfs4_op_openattr,
 		.free_res = nfs4_op_openattr_Free,
+		.resp_size = sizeof(OPENATTR4res),
 		.exp_perm_flags = EXPORT_OPTION_MD_READ_ACCESS},
 	[NFS4_OP_OPEN_CONFIRM] = {
 		.name = "OP_OPEN_CONFIRM",
 		.funct = nfs4_op_open_confirm,
 		.free_res = nfs4_op_open_confirm_Free,
+		.resp_size = sizeof(OPEN_CONFIRM4res),
 		.exp_perm_flags = 0},
 	[NFS4_OP_OPEN_DOWNGRADE] = {
 		.name = "OP_OPEN_DOWNGRADE",
 		.funct = nfs4_op_open_downgrade,
 		.free_res = nfs4_op_open_downgrade_Free,
+		.resp_size = sizeof(OPEN_DOWNGRADE4res),
 		.exp_perm_flags = EXPORT_OPTION_MD_READ_ACCESS},
 	[NFS4_OP_PUTFH] = {
 		.name = "OP_PUTFH",
 		.funct = nfs4_op_putfh,
 		.free_res = nfs4_op_putfh_Free,
+		.resp_size = sizeof(PUTFH4res),
 		.exp_perm_flags = 0},
 	[NFS4_OP_PUTPUBFH] = {
 		.name = "OP_PUTPUBFH",
 		.funct = nfs4_op_putpubfh,
 		.free_res = nfs4_op_putpubfh_Free,
+		.resp_size = sizeof(PUTPUBFH4res),
 		.exp_perm_flags = 0},
 	[NFS4_OP_PUTROOTFH] = {
 		.name = "OP_PUTROOTFH",
 		.funct = nfs4_op_putrootfh,
 		.free_res = nfs4_op_putrootfh_Free,
+		.resp_size = sizeof(PUTROOTFH4res),
 		.exp_perm_flags = 0},
 	[NFS4_OP_READ] = {
 		.name = "OP_READ",
 		.funct = nfs4_op_read,
 		.free_res = nfs4_op_read_Free,
+		.resp_size = VARIABLE_RESP_SIZE,
 		.exp_perm_flags = EXPORT_OPTION_READ_ACCESS},
 	[NFS4_OP_READDIR] = {
 		.name = "OP_READDIR",
 		.funct = nfs4_op_readdir,
 		.free_res = nfs4_op_readdir_Free,
+		.resp_size = VARIABLE_RESP_SIZE,
 		.exp_perm_flags = EXPORT_OPTION_MD_READ_ACCESS},
 	[NFS4_OP_READLINK] = {
 		.name = "OP_READLINK",
 		.funct = nfs4_op_readlink,
 		.free_res = nfs4_op_readlink_Free,
+		.resp_size = VARIABLE_RESP_SIZE,
 		.exp_perm_flags = EXPORT_OPTION_MD_READ_ACCESS},
 	[NFS4_OP_REMOVE] = {
 		.name = "OP_REMOVE",
 		.funct = nfs4_op_remove,
 		.free_res = nfs4_op_remove_Free,
+		.resp_size = sizeof(REMOVE4res),
 		.exp_perm_flags = EXPORT_OPTION_MD_WRITE_ACCESS},
 	[NFS4_OP_RENAME] = {
 		.name = "OP_RENAME",
 		.funct = nfs4_op_rename,
 		.free_res = nfs4_op_rename_Free,
+		.resp_size = sizeof(RENAME4res),
 		.exp_perm_flags = EXPORT_OPTION_MD_WRITE_ACCESS},
 	[NFS4_OP_RENEW] = {
 		.name = "OP_RENEW",
 		.funct = nfs4_op_renew,
 		.free_res = nfs4_op_renew_Free,
+		.resp_size = sizeof(RENEW4res),
 		.exp_perm_flags = 0},
 	[NFS4_OP_RESTOREFH] = {
 		.name = "OP_RESTOREFH",
 		.funct = nfs4_op_restorefh,
 		.free_res = nfs4_op_restorefh_Free,
+		.resp_size = sizeof(RESTOREFH4res),
 		.exp_perm_flags = 0},
 	[NFS4_OP_SAVEFH] = {
 		.name = "OP_SAVEFH",
 		.funct = nfs4_op_savefh,
 		.free_res = nfs4_op_savefh_Free,
+		.resp_size = sizeof(SAVEFH4res),
 		.exp_perm_flags = 0},
 	[NFS4_OP_SECINFO] = {
 		.name = "OP_SECINFO",
 		.funct = nfs4_op_secinfo,
 		.free_res = nfs4_op_secinfo_Free,
+		.resp_size = VARIABLE_RESP_SIZE,
 		.exp_perm_flags = EXPORT_OPTION_MD_READ_ACCESS},
 	[NFS4_OP_SETATTR] = {
 		.name = "OP_SETATTR",
 		.funct = nfs4_op_setattr,
 		.free_res = nfs4_op_setattr_Free,
+		.resp_size = sizeof(SETATTR4res),
 		.exp_perm_flags = EXPORT_OPTION_MD_WRITE_ACCESS},
 	[NFS4_OP_SETCLIENTID] = {
 		.name = "OP_SETCLIENTID",
 		.funct = nfs4_op_setclientid,
 		.free_res = nfs4_op_setclientid_Free,
+		.resp_size = sizeof(SETCLIENTID4res),
 		.exp_perm_flags = 0},
 	[NFS4_OP_SETCLIENTID_CONFIRM] = {
 		.name = "OP_SETCLIENTID_CONFIRM",
 		.funct = nfs4_op_setclientid_confirm,
 		.free_res = nfs4_op_setclientid_confirm_Free,
+		.resp_size = sizeof(SETCLIENTID_CONFIRM4res),
 		.exp_perm_flags = 0},
 	[NFS4_OP_VERIFY] = {
 		.name = "OP_VERIFY",
 		.funct = nfs4_op_verify,
 		.free_res = nfs4_op_verify_Free,
+		.resp_size = sizeof(VERIFY4res),
 		.exp_perm_flags = EXPORT_OPTION_MD_READ_ACCESS},
 	[NFS4_OP_WRITE] = {
 		.name = "OP_WRITE",
 		.funct = nfs4_op_write,
 		.free_res = nfs4_op_write_Free,
+		.resp_size = sizeof(WRITE4res),
 		.exp_perm_flags = EXPORT_OPTION_WRITE_ACCESS},
 	[NFS4_OP_RELEASE_LOCKOWNER] = {
 		.name = "OP_RELEASE_LOCKOWNER",
 		.funct = nfs4_op_release_lockowner,
 		.free_res = nfs4_op_release_lockowner_Free,
+		.resp_size = sizeof(RELEASE_LOCKOWNER4res),
 		.exp_perm_flags = 0},
 	[NFS4_OP_BACKCHANNEL_CTL] = {
 		.name = "OP_BACKCHANNEL_CTL",
 		.funct = nfs4_op_illegal,
 		.free_res = nfs4_op_illegal_Free,
+		.resp_size = sizeof(BACKCHANNEL_CTL4res),
 		.exp_perm_flags = 0	/* tbd */},
 	[NFS4_OP_BIND_CONN_TO_SESSION] = {
 		.name = "OP_BIND_CONN_TO_SESSION",
-		.funct = nfs4_op_illegal,
-		.free_res = nfs4_op_illegal_Free,
+		.funct = nfs4_op_bind_conn,
+		.free_res = nfs4_op_nfs4_op_bind_conn_Free,
+		.resp_size = sizeof(BIND_CONN_TO_SESSION4res),
 		.exp_perm_flags = 0	/* tbd */},
 	[NFS4_OP_EXCHANGE_ID] = {
 		.name = "OP_EXCHANGE_ID",
 		.funct = nfs4_op_exchange_id,
 		.free_res = nfs4_op_exchange_id_Free,
+		.resp_size = VARIABLE_RESP_SIZE,
 		.exp_perm_flags = 0},
 	[NFS4_OP_CREATE_SESSION] = {
 		.name = "OP_CREATE_SESSION",
 		.funct = nfs4_op_create_session,
 		.free_res = nfs4_op_create_session_Free,
+		.resp_size = sizeof(CREATE_SESSION4res),
 		.exp_perm_flags = 0},
 	[NFS4_OP_DESTROY_SESSION] = {
 		.name = "OP_DESTROY_SESSION",
 		.funct = nfs4_op_destroy_session,
 		.free_res = nfs4_op_reclaim_complete_Free,
+		.resp_size = sizeof(DESTROY_SESSION4res),
 		.exp_perm_flags = 0},
 	[NFS4_OP_FREE_STATEID] = {
 		.name = "OP_FREE_STATEID",
 		.funct = nfs4_op_free_stateid,
 		.free_res = nfs4_op_free_stateid_Free,
+		.resp_size = sizeof(FREE_STATEID4res),
 		.exp_perm_flags = 0},
 	[NFS4_OP_GET_DIR_DELEGATION] = {
 		.name = "OP_GET_DIR_DELEGATION",
 		.funct = nfs4_op_illegal,
 		.free_res = nfs4_op_illegal_Free,
+		.resp_size = sizeof(GET_DIR_DELEGATION4res),
 		.exp_perm_flags = 0	/* tbd */},
 	[NFS4_OP_GETDEVICEINFO] = {
 		.name = "OP_GETDEVICEINFO",
 		.funct = nfs4_op_getdeviceinfo,
 		.free_res = nfs4_op_getdeviceinfo_Free,
+		.resp_size = VARIABLE_RESP_SIZE,
 		.exp_perm_flags = 0},
 	[NFS4_OP_GETDEVICELIST] = {
 		.name = "OP_GETDEVICELIST",
 		.funct = nfs4_op_getdevicelist,
 		.free_res = nfs4_op_getdevicelist_Free,
+		.resp_size = VARIABLE_RESP_SIZE,
 		.exp_perm_flags = EXPORT_OPTION_MD_READ_ACCESS},
 	[NFS4_OP_LAYOUTCOMMIT] = {
 		.name = "OP_LAYOUTCOMMIT",
 		.funct = nfs4_op_layoutcommit,
 		.free_res = nfs4_op_reclaim_complete_Free,
+		.resp_size = sizeof(LAYOUTCOMMIT4res),
 		.exp_perm_flags = 0},
 	[NFS4_OP_LAYOUTGET] = {
 		.name = "OP_LAYOUTGET",
 		.funct = nfs4_op_layoutget,
 		.free_res = nfs4_op_reclaim_complete_Free,
+		.resp_size = VARIABLE_RESP_SIZE,
 		.exp_perm_flags = EXPORT_OPTION_MD_READ_ACCESS},
 	[NFS4_OP_LAYOUTRETURN] = {
 		.name = "OP_LAYOUTRETURN",
 		.funct = nfs4_op_layoutreturn,
 		.free_res = nfs4_op_reclaim_complete_Free,
+		.resp_size = sizeof(LAYOUTRETURN4res),
 		.exp_perm_flags = 0},
 	[NFS4_OP_SECINFO_NO_NAME] = {
 		.name = "OP_SECINFO_NO_NAME",
 		.funct = nfs4_op_secinfo_no_name,
 		.free_res = nfs4_op_secinfo_no_name_Free,
+		.resp_size = VARIABLE_RESP_SIZE,
 		.exp_perm_flags = EXPORT_OPTION_MD_READ_ACCESS},
 	[NFS4_OP_SEQUENCE] = {
 		.name = "OP_SEQUENCE",
 		.funct = nfs4_op_sequence,
 		.free_res = nfs4_op_sequence_Free,
+		.resp_size = sizeof(SEQUENCE4res),
 		.exp_perm_flags = 0},
 	[NFS4_OP_SET_SSV] = {
 		.name = "OP_SET_SSV",
 		.funct = nfs4_op_set_ssv,
-		.free_res = nfs4_op_reclaim_complete_Free,
+		.free_res = nfs4_op_set_ssv_Free,
+		.resp_size = sizeof(SET_SSV4res),
 		.exp_perm_flags = 0},
 	[NFS4_OP_TEST_STATEID] = {
 		.name = "OP_TEST_STATEID",
 		.funct = nfs4_op_test_stateid,
 		.free_res = nfs4_op_test_stateid_Free,
+		.resp_size = sizeof(TEST_STATEID4res),
 		.exp_perm_flags = 0},
 	[NFS4_OP_WANT_DELEGATION] = {
 		.name = "OP_WANT_DELEGATION",
 		.funct = nfs4_op_illegal,
 		.free_res = nfs4_op_illegal_Free,
+		.resp_size = sizeof(WANT_DELEGATION4res),
 		.exp_perm_flags = EXPORT_OPTION_MD_READ_ACCESS	/* tbd */},
 	[NFS4_OP_DESTROY_CLIENTID] = {
 		.name = "OP_DESTROY_CLIENTID",
 		.funct = nfs4_op_destroy_clientid,
 		.free_res = nfs4_op_destroy_clientid_Free,
+		.resp_size = sizeof(DESTROY_CLIENTID4res),
 		.exp_perm_flags = 0	/* tbd */},
 	[NFS4_OP_RECLAIM_COMPLETE] = {
 		.name = "OP_RECLAIM_COMPLETE",
 		.funct = nfs4_op_reclaim_complete,
 		.free_res = nfs4_op_reclaim_complete_Free,
+		.resp_size = sizeof(RECLAIM_COMPLETE4res),
 		.exp_perm_flags = 0},
 
 	/* NFSv4.2 */
 	[NFS4_OP_ALLOCATE] = {
-				.name = "OP_ALLOCATE",
-				.funct = nfs4_op_allocate,
-				.free_res = nfs4_op_write_Free,
-				.exp_perm_flags = 0},
+		.name = "OP_ALLOCATE",
+		.funct = nfs4_op_allocate,
+		.free_res = nfs4_op_write_Free,
+		.resp_size = sizeof(ALLOCATE4res),
+		.exp_perm_flags = 0},
 	[NFS4_OP_COPY] = {
-				.name = "OP_COPY",
-				.funct = nfs4_op_notsupp,
-				.free_res = nfs4_op_notsupp_Free,
-				.exp_perm_flags = 0},
+		.name = "OP_COPY",
+		.funct = nfs4_op_notsupp,
+		.free_res = nfs4_op_notsupp_Free,
+		.resp_size = sizeof(COPY4res),
+		.exp_perm_flags = 0},
 	[NFS4_OP_COPY_NOTIFY] = {
-				.name = "OP_COPY_NOTIFY",
-				.funct = nfs4_op_notsupp,
-				.free_res = nfs4_op_notsupp_Free,
-				.exp_perm_flags = 0},
+		.name = "OP_COPY_NOTIFY",
+		.funct = nfs4_op_notsupp,
+		.free_res = nfs4_op_notsupp_Free,
+		.resp_size = sizeof(COPY_NOTIFY4res),
+		.exp_perm_flags = 0},
 	[NFS4_OP_DEALLOCATE] = {
-				.name = "OP_DEALLOCATE",
-				.funct = nfs4_op_deallocate,
-				.free_res = nfs4_op_write_Free,
-				.exp_perm_flags = 0},
+		.name = "OP_DEALLOCATE",
+		.funct = nfs4_op_deallocate,
+		.free_res = nfs4_op_write_Free,
+		.resp_size = sizeof(DEALLOCATE4res),
+		.exp_perm_flags = 0},
 	[NFS4_OP_IO_ADVISE] = {
-				.name = "OP_IO_ADVISE",
-				.funct = nfs4_op_io_advise,
-				.free_res = nfs4_op_io_advise_Free,
-				.exp_perm_flags = 0},
+		.name = "OP_IO_ADVISE",
+		.funct = nfs4_op_io_advise,
+		.free_res = nfs4_op_io_advise_Free,
+		.resp_size = sizeof(IO_ADVISE4res),
+		.exp_perm_flags = 0},
 	[NFS4_OP_LAYOUTERROR] = {
-				.name = "OP_LAYOUTERROR",
-				.funct = nfs4_op_layouterror,
-				.free_res = nfs4_op_layouterror_Free,
-				.exp_perm_flags = 0},
+		.name = "OP_LAYOUTERROR",
+		.funct = nfs4_op_layouterror,
+		.free_res = nfs4_op_layouterror_Free,
+		.resp_size = sizeof(LAYOUTERROR4res),
+		.exp_perm_flags = 0},
 	[NFS4_OP_LAYOUTSTATS] = {
-				.name = "OP_LAYOUTSTATS",
-				.funct = nfs4_op_layoutstats,
-				.free_res = nfs4_op_layoutstats_Free,
-				.exp_perm_flags = 0},
+		.name = "OP_LAYOUTSTATS",
+		.funct = nfs4_op_layoutstats,
+		.free_res = nfs4_op_layoutstats_Free,
+		.resp_size = sizeof(LAYOUTSTATS4res),
+		.exp_perm_flags = 0},
 	[NFS4_OP_OFFLOAD_CANCEL] = {
-				.name = "OP_OFFLOAD_CANCEL",
-				.funct = nfs4_op_notsupp,
-				.free_res = nfs4_op_notsupp_Free,
-				.exp_perm_flags = 0},
+		.name = "OP_OFFLOAD_CANCEL",
+		.funct = nfs4_op_notsupp,
+		.free_res = nfs4_op_notsupp_Free,
+		.resp_size = sizeof(OFFLOAD_ABORT4res),
+		.exp_perm_flags = 0},
 	[NFS4_OP_OFFLOAD_STATUS] = {
-				.name = "OP_OFFLOAD_STATUS",
-				.funct = nfs4_op_notsupp,
-				.free_res = nfs4_op_notsupp_Free,
-				.exp_perm_flags = 0},
+		.name = "OP_OFFLOAD_STATUS",
+		.funct = nfs4_op_notsupp,
+		.free_res = nfs4_op_notsupp_Free,
+		.resp_size = sizeof(OFFLOAD_STATUS4res),
+		.exp_perm_flags = 0},
 	[NFS4_OP_READ_PLUS] = {
-				.name = "OP_READ_PLUS",
-				.funct = nfs4_op_read_plus,
-				.free_res = nfs4_op_read_plus_Free,
-				.exp_perm_flags = 0},
+		.name = "OP_READ_PLUS",
+		.funct = nfs4_op_read_plus,
+		.free_res = nfs4_op_read_plus_Free,
+		.resp_size = sizeof(READ_PLUS4res),
+		.exp_perm_flags = 0},
 	[NFS4_OP_SEEK] = {
-				.name = "OP_SEEK",
-				.funct = nfs4_op_seek,
-				.free_res = nfs4_op_write_Free,
-				.exp_perm_flags = 0},
+		.name = "OP_SEEK",
+		.funct = nfs4_op_seek,
+		.free_res = nfs4_op_write_Free,
+		.resp_size = sizeof(SEEK4res),
+		.exp_perm_flags = 0},
 	[NFS4_OP_WRITE_SAME] = {
-				.name = "OP_WRITE_SAME",
-				.funct = nfs4_op_write_same,
-				.free_res = nfs4_op_write_same_Free,
-				.exp_perm_flags = 0},
+		.name = "OP_WRITE_SAME",
+		.funct = nfs4_op_write_same,
+		.free_res = nfs4_op_write_same_Free,
+		.resp_size = sizeof(WRITE_SAME4res),
+		.exp_perm_flags = 0},
 	[NFS4_OP_CLONE] = {
-				.name = "OP_CLONE",
-				.funct = nfs4_op_notsupp,
-				.free_res = nfs4_op_notsupp_Free,
-				.exp_perm_flags = 0},
+		.name = "OP_CLONE",
+		.funct = nfs4_op_notsupp,
+		.free_res = nfs4_op_notsupp_Free,
+		.resp_size = sizeof(ILLEGAL4res),
+		.exp_perm_flags = 0},
 
 	/* NFSv4.3 */
 	[NFS4_OP_GETXATTR] = {
-				.name = "OP_GETXATTR",
-				.funct = nfs4_op_getxattr,
-				.free_res = nfs4_op_getxattr_Free,
-				.exp_perm_flags = 0},
+		.name = "OP_GETXATTR",
+		.funct = nfs4_op_getxattr,
+		.free_res = nfs4_op_getxattr_Free,
+		.resp_size = sizeof(GETXATTR4res),
+		.exp_perm_flags = 0},
 	[NFS4_OP_SETXATTR] = {
-				.name = "OP_SETXATTR",
-				.funct = nfs4_op_setxattr,
-				.free_res = nfs4_op_setxattr_Free,
-				.exp_perm_flags = 0},
+		.name = "OP_SETXATTR",
+		.funct = nfs4_op_setxattr,
+		.free_res = nfs4_op_setxattr_Free,
+		.resp_size = sizeof(SETXATTR4res),
+		.exp_perm_flags = 0},
 	[NFS4_OP_LISTXATTR] = {
-				.name = "OP_LISTXATTR",
-				.funct = nfs4_op_listxattr,
-				.free_res = nfs4_op_listxattr_Free,
-				.exp_perm_flags = 0},
+		.name = "OP_LISTXATTR",
+		.funct = nfs4_op_listxattr,
+		.free_res = nfs4_op_listxattr_Free,
+		.resp_size = sizeof(LISTXATTR4res),
+		.exp_perm_flags = 0},
 	[NFS4_OP_REMOVEXATTR] = {
-				.name = "OP_REMOVEXATTR",
-				.funct = nfs4_op_removexattr,
-				.free_res = nfs4_op_removexattr_Free,
-				.exp_perm_flags = 0},
+		.name = "OP_REMOVEXATTR",
+		.funct = nfs4_op_removexattr,
+		.free_res = nfs4_op_removexattr_Free,
+		.resp_size = sizeof(REMOVEXATTR4res),
+		.exp_perm_flags = 0},
 };
 
 /** Define the last valid NFS v4 op for each minor version.
@@ -458,6 +535,25 @@ nfs_opnum4 LastOpcode[] = {
 	NFS4_OP_REMOVEXATTR
 };
 
+void copy_tag(utf8str_cs *dest, utf8str_cs *src)
+{
+	/* Keeping the same tag as in the arguments */
+	dest->utf8string_len = src->utf8string_len;
+
+	if (dest->utf8string_len > 0) {
+
+		dest->utf8string_val = gsh_malloc(dest->utf8string_len + 1);
+
+		memcpy(dest->utf8string_val,
+		       src->utf8string_val,
+		       dest->utf8string_len);
+
+		dest->utf8string_val[dest->utf8string_len] = '\0';
+	} else {
+		dest->utf8string_val = NULL;
+	}
+}
+
 /**
  * @brief The NFS PROC4 COMPOUND
  *
@@ -493,8 +589,10 @@ int nfs4_Compound(nfs_arg_t *arg, struct svc_req *req, nfs_res_t *res)
 	nsecs_elapsed_t op_start_time;
 	struct timespec ts;
 	int perm_flags;
-	char *tagname = NULL;
 	char *notag = "NO TAG";
+	char *tagname = notag;
+	const char *bad_op_state_reason = "";
+	log_components_t alt_component = COMPONENT_NFS_V4;
 
 	if (compound4_minor > 2) {
 		LogCrit(COMPONENT_NFS_V4, "Bad Minor Version %d",
@@ -515,33 +613,33 @@ int nfs4_Compound(nfs_arg_t *arg, struct svc_req *req, nfs_res_t *res)
 	}
 
 	/* Keeping the same tag as in the arguments */
-	res->res_compound4.tag.utf8string_len =
-	    arg->arg_compound4.tag.utf8string_len;
+	copy_tag(&res->res_compound4.tag, &arg->arg_compound4.tag);
+
 	if (res->res_compound4.tag.utf8string_len > 0) {
-
-		res->res_compound4.tag.utf8string_val =
-		    gsh_malloc(res->res_compound4.tag.utf8string_len + 1);
-
-		memcpy(res->res_compound4.tag.utf8string_val,
-		       arg->arg_compound4.tag.utf8string_val,
-		       res->res_compound4.tag.utf8string_len);
-
-		res->res_compound4.tag.utf8string_val[
-			res->res_compound4.tag.utf8string_len] = '\0';
-
 		/* Check if the tag is a valid utf8 string */
 		status =
 		    nfs4_utf8string2dynamic(&(res->res_compound4.tag),
 					    UTF8_SCAN_ALL, &tagname);
 		if (status != 0) {
+			char str[LOG_BUFF_LEN];
+			struct display_buffer dspbuf = {sizeof(str), str, str};
+
+			display_opaque_bytes(
+				&dspbuf,
+				res->res_compound4.tag.utf8string_val,
+				res->res_compound4.tag.utf8string_len);
+
+			LogCrit(COMPONENT_NFS_V4,
+				"COMPOUND: bad tag %p len %d bytes %s",
+				res->res_compound4.tag.utf8string_val,
+				res->res_compound4.tag.utf8string_len,
+				str);
+
 			status = NFS4ERR_INVAL;
 			res->res_compound4.status = status;
 			res->res_compound4.resarray.resarray_len = 0;
 			return NFS_REQ_OK;
 		}
-	} else {
-		res->res_compound4.tag.utf8string_val = NULL;
-		tagname = notag;
 	}
 
 	/* Managing the operation list */
@@ -581,6 +679,9 @@ int nfs4_Compound(nfs_arg_t *arg, struct svc_req *req, nfs_res_t *res)
 	data.minorversion = compound4_minor;
 	data.req = req;
 
+	/* Initialize response size with size of compound response size. */
+	data.resp_size = sizeof(COMPOUND4res) - sizeof(nfs_resop4 *);
+
 	/* Building the client credential field */
 	if (nfs_rpc_req2client_cred(req, &(data.credential)) == -1)
 		return NFS_REQ_DROP;	/* Malformed credential */
@@ -640,66 +741,95 @@ int nfs4_Compound(nfs_arg_t *arg, struct svc_req *req, nfs_res_t *res)
 				return NFS_REQ_OK;
 			}
 		}
-
-		/* If the COMPOUND request starts with SEQUENCE, and if the
-		 * sessionids specified in SEQUENCE and DESTROY_SESSION are the
-		 * same, then DESTROY_SESSION MUST be the final operation in the
-		 * COMPOUND request.
-		 */
-		if (argarray_len > 2 && argarray[0].argop == NFS4_OP_SEQUENCE
-		    && argarray[1].argop == NFS4_OP_DESTROY_SESSION
-		    && strncmp(argarray[0].nfs_argop4_u.opsequence.sa_sessionid,
-			       argarray[1]
-				  .nfs_argop4_u.opdestroy_session.dsa_sessionid,
-			       NFS4_SESSIONID_SIZE) == 0) {
-			status = NFS4ERR_NOT_ONLY_OP;
-			res->res_compound4.status = status;
-			res->res_compound4.resarray.resarray_len = 0;
-			return NFS_REQ_OK;
-		}
 	}
 
 	for (i = 0; i < argarray_len; i++) {
 		/* Used to check if OP_SEQUENCE is the first operation */
 		data.oppos = i;
-
-		/* Verify BIND_CONN_TO_SESSION is not used in a compound
-		 * with length > 1.
-		 */
-		if (i > 0 &&
-		    argarray[i].argop == NFS4_OP_BIND_CONN_TO_SESSION) {
-			status = NFS4ERR_NOT_ONLY_OP;
-			goto bad_op_state;
-		}
-
-		/* time each op */
-		now(&ts);
-		op_start_time = timespec_diff(&ServerBootTime, &ts);
+		data.op_resp_size = sizeof(nfsstat4);
 		opcode = argarray[i].argop;
 
 		/* Handle opcode overflow */
 		if (opcode > LastOpcode[compound4_minor])
 			opcode = 0;
 
+		data.opname = optabv4[opcode].name;
+
+		LogDebug(COMPONENT_NFS_V4, "Request %d: opcode %d is %s", i,
+			 argarray[i].argop, data.opname);
+
+		/* Verify BIND_CONN_TO_SESSION is not used in a compound
+		 * with length > 1. This check is NOT redundant with the
+		 * checks above.
+		 */
+		if (i > 0 &&
+		    argarray[i].argop == NFS4_OP_BIND_CONN_TO_SESSION) {
+			status = NFS4ERR_NOT_ONLY_OP;
+			bad_op_state_reason =
+					"BIND_CONN_TO_SESSION past position 1";
+			goto bad_op_state;
+		}
+
+		/* OP_SEQUENCE is always the first operation of the request */
+		if (i > 0 && argarray[i].argop == NFS4_OP_SEQUENCE) {
+			status = NFS4ERR_SEQUENCE_POS;
+			bad_op_state_reason =
+					"SEQUENCE past position 1";
+			goto bad_op_state;
+		}
+
+		/* If a DESTROY_SESSION not the only operation, and it matches
+		 * the session specified in the SEQUENCE op (since the compound
+		 * has more than one op, we already know it MUST start with
+		 * SEQUENCE), then it MUST be the final op in the compound.
+		 */
+		if (i > 0 && argarray[i].argop == NFS4_OP_DESTROY_SESSION) {
+			bool session_compare;
+			bool bad_pos;
+
+			session_compare = memcmp(
+			    argarray[0].nfs_argop4_u.opsequence.sa_sessionid,
+			    argarray[i]
+				.nfs_argop4_u.opdestroy_session.dsa_sessionid,
+			    NFS4_SESSIONID_SIZE) == 0;
+
+			bad_pos = session_compare && i != (argarray_len - 1);
+
+			LogAtLevel(COMPONENT_SESSIONS,
+				   bad_pos ? NIV_INFO : NIV_DEBUG,
+				   "DESTROY_SESSION in position %u out of 0-%"
+				   PRIi32 " %s is %s",
+				   i, argarray_len - 1, session_compare
+					? "same session as SEQUENCE"
+					: "different session from SEQUENCE",
+				   bad_pos ? "not last op in compound" : "opk");
+
+			if (bad_pos) {
+				status = NFS4ERR_NOT_ONLY_OP;
+				bad_op_state_reason =
+				    "DESTROY_SESSION not last op in compound";
+				goto bad_op_state;
+			}
+		}
+
+		/* time each op */
+		now(&ts);
+		op_start_time = timespec_diff(&ServerBootTime, &ts);
+
 		if (compound4_minor > 0 && data.session != NULL &&
 		    data.session->fore_channel_attrs.ca_maxoperations == i) {
 			status = NFS4ERR_TOO_MANY_OPS;
+			bad_op_state_reason = "Too many operations";
 			goto bad_op_state;
 		}
 
-		LogDebug(COMPONENT_NFS_V4, "Request %d: opcode %d is %s", i,
-			 argarray[i].argop, optabv4[opcode].name);
 		perm_flags =
 		    optabv4[opcode].exp_perm_flags & EXPORT_OPTION_ACCESS_MASK;
 
 		if (perm_flags != 0) {
 			status = nfs4_Is_Fh_Empty(&data.currentFH);
 			if (status != NFS4_OK) {
-				LogDebug(COMPONENT_NFS_V4,
-					 "Status of %s for CurrentFH in position %d = %s",
-					 optabv4[opcode].name,
-					 i,
-					 nfsstat4_to_str(status));
+				bad_op_state_reason = "Empty or NULL handle";
 				goto bad_op_state;
 			}
 
@@ -723,30 +853,51 @@ int nfs4_Compound(nfs_arg_t *arg, struct svc_req *req, nfs_res_t *res)
 				else
 					status = NFS4ERR_ACCESS;
 
-				LogDebugAlt(COMPONENT_NFS_V4, COMPONENT_EXPORT,
-					    "Status of %s due to export permissions in position %d = %s",
-					    optabv4[opcode].name, i,
-					    nfsstat4_to_str(status));
- bad_op_state:
-				/* All the operation, like NFS4_OP_ACESS, have
-				 * a first replied field called .status
-				 */
-				resarray[i].nfs_resop4_u.opaccess.status =
-				    status;
-				resarray[i].resop = argarray[i].argop;
-
-				/* Do not manage the other requests in the
-				 * COMPOUND.
-				 */
-				res->res_compound4.resarray.resarray_len =
-					i + 1;
-				break;
+				bad_op_state_reason =
+						"Export permission failure";
+				alt_component = COMPONENT_EXPORT;
+				goto bad_op_state;
 			}
 		}
 
+		/* Set up the minimum/default response size and check if there
+		 * is room for it.
+		*/
+		data.op_resp_size = optabv4[opcode].resp_size;
+
+		status = check_resp_room(&data, data.op_resp_size);
+
+		if (status != NFS4_OK) {
+			bad_op_state_reason = "op response size";
+
+ bad_op_state:
+			/* Tally the response size */
+			data.resp_size += sizeof(nfs_opnum4) + sizeof(nfsstat4);
+
+			LogDebugAlt(COMPONENT_NFS_V4, alt_component,
+				    "Status of %s in position %d due to %s is %s, op response size = %"
+				    PRIu32" total response size = %"PRIu32,
+				    data.opname, i, bad_op_state_reason,
+				    nfsstat4_to_str(status),
+				    data.op_resp_size, data.resp_size);
+
+			/* All the operation, like NFS4_OP_ACCESS, have
+			 * a first replied field called .status
+			 */
+			resarray[i].nfs_resop4_u.opaccess.status = status;
+			resarray[i].resop = argarray[i].argop;
+
+			/* Do not manage the other requests in the COMPOUND. */
+			res->res_compound4.resarray.resarray_len = i + 1;
+			break;
+		}
+
+		/***************************************************************
+		 * Make the actual op call                                     *
+		 **************************************************************/
 #ifdef USE_LTTNG
 		tracepoint(nfs_rpc, v4op_start, i, argarray[i].argop,
-			   optabv4[opcode].name);
+			   data.opname);
 #endif
 
 		status = (optabv4[opcode].funct) (&argarray[i],
@@ -755,7 +906,7 @@ int nfs4_Compound(nfs_arg_t *arg, struct svc_req *req, nfs_res_t *res)
 
 #ifdef USE_LTTNG
 		tracepoint(nfs_rpc, v4op_end, i, argarray[i].argop,
-			   optabv4[opcode].name, nfsstat4_to_str(status));
+			   data.opname, nfsstat4_to_str(status));
 #endif
 
 		LogCompoundFH(&data);
@@ -767,22 +918,35 @@ int nfs4_Compound(nfs_arg_t *arg, struct svc_req *req, nfs_res_t *res)
 
 		server_stats_nfsv4_op_done(opcode, op_start_time, status);
 
-		LogDebug(COMPONENT_NFS_V4, "Status of %s in position %d = %s",
-			 optabv4[opcode].name, i, nfsstat4_to_str(status));
+		/* Tally the response size */
+		if (status != NFS4_OK &&
+		    (optabv4[opcode].resp_size != VARIABLE_RESP_SIZE ||
+		     data.op_resp_size == VARIABLE_RESP_SIZE)) {
+			/* If the op failed and has a static response size, or
+			 * it has a variable size that hasn't been set, use the
+			 * sizeof nfsstat4 instead.
+			 */
+			data.op_resp_size = sizeof(nfsstat4);
+		}
+
+		data.resp_size += sizeof(nfs_opnum4) + data.op_resp_size;
+
+		LogDebug(COMPONENT_NFS_V4,
+			 "Status of %s in position %d = %s, op response size is %"
+			 PRIu32" total response size is %"PRIu32,
+			 data.opname, i, nfsstat4_to_str(status),
+			 data.op_resp_size, data.resp_size);
 
 		if (status != NFS4_OK) {
 			/* An error occured, we do not manage the other requests
 			 * in the COMPOUND, this may be a regular behavior
 			 */
 			res->res_compound4.resarray.resarray_len = i + 1;
-
 			break;
 		}
 
-		/* Check Req size */
-
 		/* NFS_V4.1 specific stuff */
-		if (data.use_drc) {
+		if (data.use_slot_cached_result) {
 			/* Replay cache, only true for SEQUENCE or
 			 * CREATE_SESSION w/o SEQUENCE. Since will only be set
 			 * in those cases, no need to check operation or
@@ -793,11 +957,12 @@ int nfs4_Compound(nfs_arg_t *arg, struct svc_req *req, nfs_res_t *res)
 			gsh_free(res->res_compound4.resarray.resarray_val);
 
 			/* Copy the reply from the cache */
-			res->res_compound4_extended = *data.cached_res;
-			status = ((COMPOUND4res *) data.cached_res)->status;
+			res->res_compound4_extended = *data.cached_result;
+			status = ((COMPOUND4res *) data.cached_result)->status;
 			LogFullDebug(COMPONENT_SESSIONS,
 				     "Use session replay cache %p result %s",
-				     data.cached_res, nfsstat4_to_str(status));
+				     data.cached_result,
+				     nfsstat4_to_str(status));
 			break;	/* Exit the for loop */
 		}
 	}			/* for */
@@ -812,25 +977,63 @@ int nfs4_Compound(nfs_arg_t *arg, struct svc_req *req, nfs_res_t *res)
 	/* Manage session's DRC: keep NFS4.1 replay for later use, but don't
 	 * save a replayed result again.
 	 */
-	if (data.cached_res != NULL && !data.use_drc) {
+	if (data.sa_cachethis) {
 		/* Pointer has been set by nfs4_op_sequence and points to slot
 		 * to cache result in.
 		 */
 		LogFullDebug(COMPONENT_SESSIONS,
 			     "Save result in session replay cache %p sizeof nfs_res_t=%d",
-			     data.cached_res, (int)sizeof(nfs_res_t));
+			     data.cached_result, (int)sizeof(nfs_res_t));
 
 		/* Indicate to nfs4_Compound_Free that this reply is cached. */
 		res->res_compound4_extended.res_cached = true;
 
-		/* If the cache is already in use, free it. */
-		if (data.cached_res->res_cached) {
-			data.cached_res->res_cached = false;
-			nfs4_Compound_Free((nfs_res_t *) data.cached_res);
+		/* Save the result in the cache (copy out of the result array
+		 * into the slot cache (which is pointed to by
+		 * data.cached_result).
+		 */
+		*data.cached_result = res->res_compound4_extended;
+	} else if (compound4_minor > 0 && !data.use_slot_cached_result &&
+		   argarray[0].argop == NFS4_OP_SEQUENCE &&
+		   data.cached_result != NULL) {
+		/* We need to cache an "uncached" response. The length is
+		 * 1 if only one op processed, otherwise 2. */
+		struct COMPOUND4res *c_res = &data.cached_result->res_compound4;
+		u_int resarray_len =
+			res->res_compound4.resarray.resarray_len == 1 ? 1 : 2;
+		struct nfs_resop4 *res0;
+
+		c_res->resarray.resarray_len = resarray_len;
+		c_res->resarray.resarray_val =
+			gsh_calloc(resarray_len, sizeof(struct nfs_resop4));
+		copy_tag(&c_res->tag, &res->res_compound4.tag);
+		res0 = c_res->resarray.resarray_val;
+
+		/* Copy the sequence result. */
+		*res0 = res->res_compound4.resarray.resarray_val[0];
+		c_res->status = res0->nfs_resop4_u.opillegal.status;
+
+		if (resarray_len == 2) {
+			struct nfs_resop4 *res1 = res0 + 1;
+
+			/* Shallow copy response since we will override any
+			 * resok or any negative response that might have
+			 * allocated data.
+			 */
+			*res1 = res->res_compound4.resarray.resarray_val[1];
+
+			if (res1->nfs_resop4_u.opillegal.status == NFS4_OK ||
+			    res1->nfs_resop4_u.opillegal.status ==
+							NFS4ERR_DENIED) {
+				res1->nfs_resop4_u.opillegal.status =
+						NFS4ERR_RETRY_UNCACHED_REP;
+			}
+
+			c_res->status = res1->nfs_resop4_u.opillegal.status;
 		}
 
-		/* Save the result in the cache. */
-		*data.cached_res = res->res_compound4_extended;
+		/* Indicate that this reply is cached in slot cache. */
+		data.cached_result->res_cached = true;
 	}
 
 	/* If we have reserved a lease, update it and release it */
@@ -849,6 +1052,13 @@ int nfs4_Compound(nfs_arg_t *arg, struct svc_req *req, nfs_res_t *res)
 
 	compound_data_Free(&data);
 
+	/* release current active export in op_ctx. */
+	if (op_ctx->ctx_export) {
+		put_gsh_export(op_ctx->ctx_export);
+		op_ctx->ctx_export = NULL;
+		op_ctx->fsal_export = NULL;
+	}
+
 	return NFS_REQ_OK;
 }				/* nfs4_Compound */
 
@@ -911,9 +1121,10 @@ void nfs4_Compound_Free(nfs_res_t *res)
 	}
 
 	gsh_free(res->res_compound4.resarray.resarray_val);
+	res->res_compound4.resarray.resarray_val = NULL;
 
-	if (res->res_compound4.tag.utf8string_val)
-		gsh_free(res->res_compound4.tag.utf8string_val);
+	gsh_free(res->res_compound4.tag.utf8string_val);
+	res->res_compound4.tag.utf8string_val = NULL;
 }
 
 /**
@@ -931,17 +1142,18 @@ void compound_data_Free(compound_data_t *data)
 	set_saved_entry(data, NULL);
 
 	if (data->session) {
+		if (data->slot != UINT32_MAX) {
+			nfs41_session_slot_t *slot;
+
+			/* Release the slot if in use */
+			slot = &data->session->fc_slots[data->slot];
+			PTHREAD_MUTEX_unlock(&slot->lock);
+		}
+
 		dec_session_ref(data->session);
 		data->session = NULL;
 	}
 
-	/* Release CurrentFH reference to export. */
-	if (op_ctx->ctx_export) {
-		put_gsh_export(op_ctx->ctx_export);
-		op_ctx->ctx_export = NULL;
-		op_ctx->fsal_export = NULL;
-	}
-
 	/* Release SavedFH reference to export. */
 	if (data->saved_export) {
 		put_gsh_export(data->saved_export);
